/*
 * vctest.cpp
 * Copyright (C) 2008  ~V Ўu
 *
 * ̃vOɂ GNU GPLv2 Kp܂B
 */

/*
 * to compile:
 *   cl /EHsc vctest.cpp vfw32.lib
 */

#include <stdio.h>
#include <vector>
#include <algorithm>

#include <windows.h>
#include <vfw.h>

using namespace std;

DWORD dwProcessAffinityMask, dwSystemAffinityMask;

void FCC2String(char *buf, DWORD fcc)
{
	sprintf(buf, "%c%c%c%c", (fcc&0xff), ((fcc>>8)&0xff), ((fcc>>16)&0xff), (fcc>>24));
}

HANDLE hFile;
vector<pair<LARGE_INTEGER, DWORD> > aviindex;
vector<double> enctime;
vector<double> dectime;
double totalenctime;
double totaldectime;
LARGE_INTEGER liTotalOrigSize;
LARGE_INTEGER liTotalEncSize;


DWORD ScanSubChunk(void)
{
	DWORD name;
	DWORD len;
	DWORD cbRead;

	ReadFile(hFile, &name, 4, &cbRead, NULL);
	if (cbRead != 4)
		return 0;
	ReadFile(hFile, &len, 4, &cbRead, NULL);
	if (cbRead != 4)
		return 0;

	//printf("%08X %d\n", name, len);
	if (name == MAKEFOURCC('R','I','F','F') || name == MAKEFOURCC('L','I','S','T'))
	{
		DWORD type;
		DWORD curlen;

		ReadFile(hFile, &type, 4, &cbRead, NULL);
		if (cbRead != 4)
			return 8;

		curlen = 4;
		while (curlen < len)
			curlen += ScanSubChunk();
	}
	else
	{
		LARGE_INTEGER liZero;
		LARGE_INTEGER liLen;
		LARGE_INTEGER liCurPos;

		liZero.QuadPart = 0;
		liLen.QuadPart = ((len + 1) / 2) * 2;
		SetFilePointerEx(hFile, liZero, &liCurPos, FILE_CURRENT);
		SetFilePointerEx(hFile, liLen, NULL, FILE_CURRENT);
		if ((name & 0x0000ffff) == 0x3030)
			aviindex.push_back(pair<LARGE_INTEGER, DWORD>(liCurPos, len));
		len = liLen.LowPart;
	}
	return len + 8;
}

void ScanChunk(void)
{
	while (ScanSubChunk() > 0)
		/* NOTHING */;
}

char *pRandomBuf;

void InitFlushCache(void)
{
	pRandomBuf = (char *)VirtualAlloc(NULL, 16*1024*1024, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	for (int i = 0; i < 16*1024*1024; i++)
		pRandomBuf[i] = rand();
}

void FlushCache(void)
{
	char ch;
	for (DWORD mask = 1; mask != 0; mask <<= 1)
	{
		if (SetThreadAffinityMask(GetCurrentThread(), mask) != 0)
		{
			for (int i = 0; i < 16*1024*1024; i+=32)
				ch = pRandomBuf[i];
		}
	}
	SetThreadAffinityMask(GetCurrentThread(), dwProcessAffinityMask);
}

int main(int argc, char **argv)
{
	PAVISTREAM pStream;
	union
	{
		BITMAPINFOHEADER *pbmihOrig;
		void *pbmihOrigBuf;
	};
	union
	{
		BITMAPINFOHEADER *pbmihEncoded;
		void *pbmihEncodedBuf;
	};
	union
	{
		BITMAPINFOHEADER *pbmihDecoded;
		void *pbmihDecodedBuf;
	};
	LONG cbFormatOrig;
	LONG cbFormatEncoded;
	HIC hicCompress;
	HIC hicDecompress;
	DWORD dw;
	__declspec(align(4096)) static char bufOrig[1024*1024*4];
	__declspec(align(4096)) static char bufEncoded[1024*1024*4];
	__declspec(align(4096)) static char bufDecoded[1024*1024*4];
	LONG cbOrig;
	HRESULT hr;
	LARGE_INTEGER liStartEncode, liEndEncode;
	LARGE_INTEGER liStartDecode, liEndDecode;
	LARGE_INTEGER liFreq;
	DWORD dwAviIndexFlag;
	char buf[256];
	COMPVARS cv;
	BOOL b;
	DWORD dwHandler;
	AVISTREAMINFO asi;
	DWORD cbRead;
	DWORD cbState;
	void *pState;
	BOOL bCheckLossless;

	if (argc <= 1)
	{
		printf("usage: %s <AVI file name> [-c]\n", argv[0]);
		return 0;
	}

	SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);

	if (argc > 2 && strcmp(argv[2], "-c") == 0)
		bCheckLossless = 1;
	else
		bCheckLossless = 0;

	liTotalOrigSize.QuadPart = 0;
	liTotalEncSize.QuadPart = 0;

	InitFlushCache();
	GetProcessAffinityMask(GetCurrentProcess(), &dwProcessAffinityMask, &dwSystemAffinityMask);
	QueryPerformanceFrequency(&liFreq);
	AVIFileInit();

	AVIStreamOpenFromFile(&pStream, argv[1], streamtypeVIDEO, 0, OF_SHARE_DENY_WRITE, NULL);
	hr = AVIStreamReadFormat(pStream, 0, NULL, &cbFormatOrig);
	if (FAILED(hr)) { printf("AVIStreamReadFormat() failed: %08X\n", hr); }
	pbmihOrigBuf = malloc(cbFormatOrig);
	hr = AVIStreamReadFormat(pStream, 0, pbmihOrig, &cbFormatOrig);
	if (FAILED(hr)) { printf("AVIStreamReadFormat() failed: %08X\n", hr); }
	AVIStreamInfo(pStream, &asi, sizeof(asi));
	printf("dwLength = %d\ndwScale=%d\ndwRate=%d\n", asi.dwLength, asi.dwScale, asi.dwRate);

	hFile = CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	ScanChunk();

	printf("Lossless Checking is %s\n", bCheckLossless ? "enabled" : "disabled");

	FCC2String(buf, pbmihOrig->biCompression);
	printf("%s\n", buf);

	memset(&cv, 0, sizeof(cv));
	cv.cbSize = sizeof(cv);
	b = ICCompressorChoose(NULL, 0, pbmihOrig, NULL, &cv, "قق");
	if (!b)
	{
		printf("ICCompressorChoose() returned FALSE\n");
		return 0;
	}
	dwHandler = cv.fccHandler;
	cbState = ICGetStateSize(cv.hic);
	pState = malloc(cbState+1);
	ICGetState(cv.hic, pState, cbState);
	ICCompressorFree(&cv);
	printf("statesize=%d\n", cbState);

	FCC2String(buf, dwHandler);
	printf("%s\n", buf);

	hicCompress = ICOpen(ICTYPE_VIDEO, dwHandler, ICMODE_COMPRESS);
	if (hicCompress == NULL) { printf("ICOpen() failed\n"); }
	dw = ICSetState(hicCompress, pState, cbState);
	if (dw == 0) { printf("ICSetState() failed\n"); }
	cbFormatEncoded = ICCompressGetFormatSize(hicCompress, pbmihOrig);
	pbmihEncodedBuf = malloc(cbFormatEncoded);
	dw = ICCompressGetFormat(hicCompress, pbmihOrig, pbmihEncoded);
	if (dw != ICERR_OK) { printf("ICCompressGetFormat() failed  dw=%08x\n", dw); }

	hicDecompress = ICOpen(ICTYPE_VIDEO, dwHandler, ICMODE_DECOMPRESS);
	if (hicDecompress == NULL) { printf("ICOpen() failed\n"); }
	pbmihDecodedBuf = malloc(cbFormatOrig);
	memcpy(pbmihDecoded, pbmihOrig, cbFormatOrig);

	dw = ICCompressBegin(hicCompress, pbmihOrig, pbmihEncoded);
	if (dw != ICERR_OK) { printf("ICCompressBegin() failed  dw=%08x\n", dw); }
	dw = ICDecompressBegin(hicDecompress, pbmihEncoded, pbmihDecoded);
	if (dw != ICERR_OK) { printf("ICDecompressBegin() failed  dw=%08x\n", dw); }

	for (int i = 0; i < asi.dwLength; i ++)
	{
		double etime, dtime;
		//hr = AVIStreamRead(pStream, i, 1, bufOrig, sizeof(bufOrig), &cbOrig, NULL);
		//if (FAILED(hr)) { printf("AVIStreamRead() failed: %08X\n", hr); }
		pair<LARGE_INTEGER, DWORD> idx = aviindex[i];
		b = SetFilePointerEx(hFile, idx.first, NULL, FILE_BEGIN);
		if (!b) { printf("SetFilePointerEx() failed  GetLastError=%08x\n", GetLastError()); }
		b = ReadFile(hFile, bufOrig, idx.second, &cbRead, NULL);
		if (!b) { printf("ReadFile() failed  GetLastError=%08x\n", GetLastError()); }
		FlushCache();
		QueryPerformanceCounter(&liStartEncode);
		dw = ICCompress(hicCompress, i == 0 ? ICCOMPRESS_KEYFRAME : 0, pbmihEncoded, bufEncoded, pbmihOrig, bufOrig, NULL, &dwAviIndexFlag, i, sizeof(bufEncoded), 0, NULL, NULL);
		QueryPerformanceCounter(&liEndEncode);
		if (dw != ICERR_OK) { printf("ICCompress() failed  dw=%08x\n", dw); }
		FlushCache();
		QueryPerformanceCounter(&liStartDecode);
		dw = ICDecompress(hicDecompress, ((dwAviIndexFlag&AVIIF_KEYFRAME) ? 0 : ICDECOMPRESS_NOTKEYFRAME), pbmihEncoded, bufEncoded, pbmihDecoded, bufDecoded);
		QueryPerformanceCounter(&liEndDecode);
		if (dw != ICERR_OK) { printf("ICDecompress() failed  dw=%08x\n", dw); }
		etime = (liEndEncode.QuadPart - liStartEncode.QuadPart) / (double)liFreq.QuadPart * 1000.0;
		dtime = (liEndDecode.QuadPart - liStartDecode.QuadPart) / (double)liFreq.QuadPart * 1000.0;
		printf("%5d/%5d %10.6fms %10.6fms %7dbytes (%4.1f%%) %s\n",
			i, asi.dwLength, etime, dtime,
			pbmihEncoded->biSizeImage, (double)pbmihEncoded->biSizeImage / (double)pbmihOrig->biSizeImage * 100.0,
			(pbmihEncoded->biSizeImage == 0) ? "NUL" : (dwAviIndexFlag&AVIIF_KEYFRAME) ? "KEY" : "");
		if (bCheckLossless && memcmp(bufOrig, bufDecoded, pbmihOrig->biSizeImage) != 0)
		{
			printf("NOT LOSSLESS!!! - aborting\n");
			break;
		}
		liTotalOrigSize.QuadPart += pbmihOrig->biSizeImage;
		liTotalEncSize.QuadPart += pbmihEncoded->biSizeImage;
		enctime.push_back(etime);
		dectime.push_back(dtime);
		totalenctime += etime;
		totaldectime += dtime;
	}
	ICCompressEnd(hicCompress);
	ICDecompressEnd(hicDecompress);

	ICClose(hicCompress);
	ICClose(hicDecompress);
	AVIStreamRelease(pStream);

	AVIFileExit();

	printf("\n");

	printf("Size: %I64d/%I64d (%4.1f%%, %4.2f)\n",
		liTotalEncSize.QuadPart, liTotalOrigSize.QuadPart,
		(double)liTotalEncSize.QuadPart/(double)liTotalOrigSize.QuadPart*100.0,
		(double)liTotalOrigSize.QuadPart/(double)liTotalEncSize.QuadPart);

	sort(enctime.begin(), enctime.end());
	sort(dectime.begin(), dectime.end());

	printf("Encode time: %fms/%df = %fms/f\n", totalenctime, asi.dwLength, totalenctime/asi.dwLength);
	printf("    min  %f\n", enctime[0]);
	printf("    10%%  %f\n", enctime[enctime.size()*0.10]);
	printf("    25%%  %f\n", enctime[enctime.size()*0.25]);
	printf("    50%%  %f\n", enctime[enctime.size()*0.50]);
	printf("    75%%  %f\n", enctime[enctime.size()*0.75]);
	printf("    90%%  %f\n", enctime[enctime.size()*0.90]);
	printf("    max  %f\n", enctime[enctime.size()-1]);

	printf("Decode time: %fms/%df = %fms/f\n", totaldectime, asi.dwLength, totaldectime/asi.dwLength);
	printf("    min  %f\n", dectime[0]);
	printf("    10%%  %f\n", dectime[dectime.size()*0.10]);
	printf("    25%%  %f\n", dectime[dectime.size()*0.25]);
	printf("    50%%  %f\n", dectime[dectime.size()*0.50]);
	printf("    75%%  %f\n", dectime[dectime.size()*0.75]);
	printf("    90%%  %f\n", dectime[dectime.size()*0.90]);
	printf("    max  %f\n", dectime[dectime.size()-1]);
	return 0;
}
